/*-
 * Copyright (C) @BABOLO  2002 Oct 05
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef lint
char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Oct 05\n\
@(#)All rights reserved.\n";
static char rcsid[] = "$Id: ip_extend.c,v 1.1 2002/12/08 18:22:52 babolo Exp $";
#endif /* not lint */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <postgres.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fmgr.h>
#include <utils/inet.h>
#include <nodes/nodes.h>
#include <nodes/execnodes.h>
#define ip_v4addr(inetptr)   (((inet_struct *)VARDATA(inetptr))->addr.ipv4_addr)
#define ip_bits(inetptr)     (((inet_struct *)VARDATA(inetptr))->bits)
#define ip_family(inetptr)   (((inet_struct *)VARDATA(inetptr))->family)
#define ip_type(inetptr)     (((inet_struct *)VARDATA(inetptr))->type)
#define ip_addrsize(inetptr) (((inet_struct *)VARDATA(inetptr))->family == AF_INET ? 4 : -1)
#define INET_MEMSIZE(a) (VARHDRSZ + ((char*)&ip_v4addr(a) - (char*)VARDATA(a)) + ip_addrsize(a))
#define result ((results*)fcinfo->flinfo->fn_extra)

typedef struct {
    u_long current;
    u_long brodcast;
    char   done;
} results;

PG_FUNCTION_INFO_V1(ip_set);
Datum
ip_set(PG_FUNCTION_ARGS) {
    int i;
    u_long mask;
    u_long step;
    inet *src, *dst;
    MemoryContext oldcontext;

    if  (PG_ARGISNULL(0) || PG_ARGISNULL(1)) {
        ((ReturnSetInfo *) fcinfo->resultinfo)->isDone = ExprEndResult;
        PG_RETURN_NULL();
    }
    if  (PG_GETARG_INT32(1) > 32 || PG_GETARG_INT32(1) < 0)        elog(ERROR, "ip_set: Out of range");
    if  (fcinfo->resultinfo == NULL || !IsA(fcinfo->resultinfo, ReturnSetInfo))
                   elog(ERROR, "ip_set: function called in context that does not accept a set result");

    src = PG_GETARG_INET_P(0);
    if  (ip_family(src) != AF_INET) elog(ERROR, "ip_set: unknown address family (%d)", ip_family(src));
    if  (ip_bits(src) > PG_GETARG_INT32(1)) {
        ((ReturnSetInfo *) fcinfo->resultinfo)->isDone = ExprEndResult;
        PG_RETURN_NULL();
    }
    step = 1 << (32 - PG_GETARG_INT32(1));
    if  (fcinfo->flinfo->fn_extra == NULL || result == NULL || result->done) {
        mask = 0;
        for (i = 32 - ip_bits(src); i > 0; i--) mask = (mask << 1) | 1;
        oldcontext = MemoryContextSwitchTo(fcinfo->flinfo->fn_mcxt);
        result = (results*)palloc(sizeof(results));
        if  (result == NULL)                                         elog(ERROR, "ip_set: memory low");
	MemSet(result, 0, sizeof(results));
        result->current = ~mask & ntohl(ip_v4addr(src));
        result->brodcast = mask | ntohl(ip_v4addr(src));
        result->done = 0;
        MemoryContextSwitchTo(oldcontext);
    }
    if  (result->current <= result->brodcast) {
        dst = (inet*)palloc(VARHDRSZ + sizeof(inet_struct));
        if  (dst == NULL)                                            elog(ERROR, "ip_set: memory low");
        MemSet(dst, 0, VARHDRSZ + sizeof(inet_struct));
        ip_v4addr(dst) = htonl(result->current);
        ip_family(dst) = AF_INET;
        ip_bits(dst) = PG_GETARG_INT32(1);
        ip_type(dst) = 0;
        VARATT_SIZEP(dst) = INET_MEMSIZE(dst);
        ((ReturnSetInfo *) fcinfo->resultinfo)->isDone = ExprSingleResult; /* ExprMultipleResult (?) */
        result->current += step;
        PG_RETURN_INET_P(dst);
    }else{
        ((ReturnSetInfo *) fcinfo->resultinfo)->isDone = ExprEndResult;
        result->done = 1;
        PG_RETURN_NULL();
}   }

PG_FUNCTION_INFO_V1(ip_supernet);
Datum
ip_supernet(PG_FUNCTION_ARGS) {
    inet *src, *dst;
    MemoryContext oldcontext;

    if  (PG_ARGISNULL(0)) {
        ((ReturnSetInfo *) fcinfo->resultinfo)->isDone = ExprEndResult;
        PG_RETURN_NULL();
    }
    if  (fcinfo->resultinfo == NULL || !IsA(fcinfo->resultinfo, ReturnSetInfo))
              elog(ERROR, "ip_supernet: function called in context that does not accept a set result");

    src = PG_GETARG_INET_P(0);
    if  (ip_family(src) != AF_INET)
                               elog(ERROR, "ip_supernet: unknown address family (%d)", ip_family(src));
    if  ( fcinfo->flinfo->fn_extra == NULL || result == NULL || result->done < 0) {
        oldcontext = MemoryContextSwitchTo(fcinfo->flinfo->fn_mcxt);
        result = (results*)palloc(sizeof(results));
        if  (result == NULL)                                    elog(ERROR, "ip_supernet: memory low");
	MemSet(result, 0, sizeof(results));
        result->current = ip_v4addr(src);
        result->done = ip_bits(src);
        MemoryContextSwitchTo(oldcontext);
    }
    if  (--(result->done) >= 0) {
        dst = (inet*)palloc(VARHDRSZ + sizeof(inet_struct));
        if  (dst == NULL)                                       elog(ERROR, "ip_supernet: memory low");
        MemSet(dst, 0, VARHDRSZ + sizeof(inet_struct));
        ip_v4addr(dst) = result->current;
        ip_family(dst) = AF_INET;
        ip_bits(dst) = result->done;
        ip_type(dst) = 0;
                                                      /* ExprMultipleResult (?) */
        ((ReturnSetInfo *) fcinfo->resultinfo)->isDone = ExprSingleResult;
        VARATT_SIZEP(dst) = INET_MEMSIZE(dst);
        PG_RETURN_INET_P(dst);
    }else{
        ((ReturnSetInfo *) fcinfo->resultinfo)->isDone = ExprEndResult;
        PG_RETURN_NULL();
}   }

PG_FUNCTION_INFO_V1(int4_inet);
Datum
int4_inet(PG_FUNCTION_ARGS) {
    inet *dst;

    if  (PG_ARGISNULL(0)) PG_RETURN_NULL();
    dst = (inet*)palloc(VARHDRSZ + sizeof(inet_struct));
    if  (dst == NULL)                                             elog(ERROR, "int4_inet: memory low");
    MemSet(dst, 0, VARHDRSZ + sizeof(inet_struct));
    ip_v4addr(dst) = htonl(PG_GETARG_INT32(0));
    ip_family(dst) = AF_INET;
    ip_bits(dst) = 32;
    ip_type(dst) = 0;
    VARATT_SIZEP(dst) = INET_MEMSIZE(dst);
    PG_RETURN_INET_P(dst);
};

PG_FUNCTION_INFO_V1(int8_inet);
Datum
int8_inet(PG_FUNCTION_ARGS) {
    inet *dst;

    if  (PG_ARGISNULL(0)) PG_RETURN_NULL();
    dst = (inet*)palloc(VARHDRSZ + sizeof(inet_struct));
    if  (dst == NULL)                                             elog(ERROR, "int8_inet: memory low");
    MemSet(dst, 0, VARHDRSZ + sizeof(inet_struct));
    ip_v4addr(dst) = htonl((u_long)PG_GETARG_INT64(0));
    ip_family(dst) = AF_INET;
    ip_bits(dst) = 32;
    ip_type(dst) = 0;
    VARATT_SIZEP(dst) = INET_MEMSIZE(dst);
    PG_RETURN_INET_P(dst);
};

PG_FUNCTION_INFO_V1(inet_int4);
Datum
inet_int4(PG_FUNCTION_ARGS) {
    u_long n;
    inet *src;

    if  (PG_ARGISNULL(0)) PG_RETURN_NULL();
    src = PG_GETARG_INET_P(0);
    if  (ip_family(src) == AF_INET) {
        PG_RETURN_INT32(ntohl(ip_v4addr(src)));
    }else{
        elog(ERROR, "unknown address family (%d)", ip_family(src));
        PG_RETURN_NULL();
};  };

PG_FUNCTION_INFO_V1(inet_int8);
Datum
inet_int8(PG_FUNCTION_ARGS) {
    u_long n;
    inet *src;

    if  (PG_ARGISNULL(0)) PG_RETURN_NULL();
    src = PG_GETARG_INET_P(0);
    if  (ip_family(src) == AF_INET) {
        PG_RETURN_INT64(ntohl(ip_v4addr(src)));
    }else{
        elog(ERROR, "unknown address family (%d)", ip_family(src));
        PG_RETURN_NULL();
};  };

PG_FUNCTION_INFO_V1(inet_type);
Datum
inet_type(PG_FUNCTION_ARGS) {
    u_long n;
    inet *src;

    if  (PG_ARGISNULL(0)) PG_RETURN_NULL();
    src = PG_GETARG_INET_P(0);
    if  (ip_family(src) == AF_INET) {
        PG_RETURN_INT16(ip_type(src));
    }else{
        elog(ERROR, "unknown address family (%d)", ip_family(src));
        PG_RETURN_NULL();
};  };
